# -*- coding: utf-8 -*-
"""
Dummy backend for unsupported platforms.
"""

from __future__ import annotations

# system imports
import uuid

# local imports
from .base import Notification, DesktopNotifierBase


class DummyNotificationCenter(DesktopNotifierBase):
    """A dummy backend for unsupported platforms"""

    def __init__(
        self,
        app_name: str = "Python",
        notification_limit: int | None = None,
    ) -> None:
        super().__init__(app_name, notification_limit)

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        return True

    async def has_authorisation(self) -> bool:
        """
        Whether we have authorisation to send notifications.
        """
        return True

    async def _send(
        self,
        notification: Notification,
        notification_to_replace: Notification | None,
    ) -> str:
        if notification_to_replace:
            return str(notification_to_replace.identifier)
        else:
            return str(uuid.uuid4())

    async def _clear(self, notification: Notification) -> None:
        pass

    async def _clear_all(self) -> None:
        pass



# -*- coding: utf-8 -*-
"""
Support module for macOS.

"""

import platform
import ctypes

from typing import cast

from packaging.version import Version
from rubicon.objc import ObjCClass
from rubicon.objc.runtime import load_library


macos_version = Version(platform.mac_ver()[0])


__all__ = [
    "is_bundle",
    "is_signed_bundle",
    "macos_version",
]


sec = load_library("Security")
foundation = load_library("Foundation")

NSBundle = ObjCClass("NSBundle")

# Security/SecRequirement.h

kSecCSDefaultFlags = 0
kSecCSCheckAllArchitectures = 1 << 0
kSecCSDoNotValidateExecutable = 1 << 1
kSecCSDoNotValidateResources = 1 << 2
kSecCSCheckNestedCode = 1 << 3
kSecCSStrictValidate = 1 << 4


def is_bundle() -> bool:
    """
    Detect if we are in an app bundle / framework.

    :returns: Whether we are inside a valid app bundle or framework.
    """

    main_bundle = NSBundle.mainBundle

    return main_bundle.bundleIdentifier is not None


def is_signed_bundle() -> bool:
    """
    Detect if we are in a signed app bundle / framework.

    :returns: Whether we are inside a signed app bundle or framework.
    """

    main_bundle = NSBundle.mainBundle

    if main_bundle.bundleIdentifier is None:
        return False

    # Check for valid signature.

    static_code = ctypes.c_void_p(0)

    err = sec.SecStaticCodeCreateWithPath(
        main_bundle.bundleURL, kSecCSDefaultFlags, ctypes.byref(static_code)
    )

    if err != 0:
        return False

    signed_status = sec.SecStaticCodeCheckValidityWithErrors(
        static_code,
        kSecCSCheckAllArchitectures | kSecCSCheckNestedCode | kSecCSStrictValidate,
        None,
        None,
    )

    return cast(int, signed_status) == 0



# -*- coding: utf-8 -*-
"""
Notification backend for Windows. Unlike other platforms, sending rich "toast"
notifications cannot be done via FFI / ctypes because the C winapi only supports basic
notifications with a title and message. This backend therefore requires interaction with
the Windows Runtime and uses the winrt package with compiled components published by
Microsoft (https://github.com/microsoft/xlang, https://pypi.org/project/winrt/).
"""

from __future__ import annotations

# system imports
import uuid
import logging
from xml.etree.ElementTree import Element, SubElement, tostring
from typing import TypeVar, Any, cast

# external imports
import winreg
from winsdk.windows.ui.notifications import (
    ToastNotificationManager,
    ToastNotificationPriority,
    NotificationSetting,
    ToastNotification,
    ToastActivatedEventArgs,
    ToastDismissalReason,
)
from winsdk.windows.data.xml import dom
from winsdk.windows.applicationmodel.core import CoreApplication
from winsdk.windows.foundation import IPropertyValue, PropertyType
import winsdk._winrt as _winrt

# local imports
from .base import Notification, DesktopNotifierBase, Urgency


__all__ = ["WinRTDesktopNotifier"]

logger = logging.getLogger(__name__)

T = TypeVar("T")


def register_hkey(app_id: str, app_name: str) -> None:
    winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)  # type:ignore
    key_path = f"SOFTWARE\\Classes\\AppUserModelId\\{app_id}"
    with winreg.CreateKeyEx(  # type:ignore
        winreg.HKEY_CURRENT_USER, key_path  # type:ignore
    ) as master_key:
        winreg.SetValueEx(  # type:ignore
            master_key, "DisplayName", 0, winreg.REG_SZ, app_name  # type:ignore
        )


class WinRTDesktopNotifier(DesktopNotifierBase):
    """Notification backend for the Windows Runtime

    :param app_name: The name of the app. This has no effect since the app name will be
        automatically determined.
    :param notification_limit: Maximum number of notifications to keep in the system's
        notification center.
    """

    _to_native_urgency = {
        Urgency.Low: ToastNotificationPriority.DEFAULT,
        Urgency.Normal: ToastNotificationPriority.DEFAULT,
        Urgency.Critical: ToastNotificationPriority.HIGH,
    }

    def __init__(
        self,
        app_name: str = "Python",
        notification_limit: int | None = None,
    ) -> None:
        super().__init__(app_name, notification_limit)
        self.manager = ToastNotificationManager.get_default()

        # Prefer using the real App ID if detected, fall back to user-provided name
        # and icon otherwise.
        if CoreApplication.id != "":
            self.app_id = CoreApplication.id
        else:
            self.app_id = app_name
            register_hkey(app_name, app_name)
        self.notifier = self.manager.create_toast_notifier(self.app_id)

    async def request_authorisation(self) -> bool:
        """
        Request authorisation to send notifications.

        :returns: Whether authorisation has been granted.
        """
        return bool(self.notifier.setting == NotificationSetting.ENABLED)

    async def has_authorisation(self) -> bool:
        """
        Whether we have authorisation to send notifications.
        """
        return bool(self.notifier.setting == NotificationSetting.ENABLED)

    async def _send(
        self,
        notification: Notification,
        notification_to_replace: Notification | None,
    ) -> str:
        """
        Asynchronously sends a notification.

        :param notification: Notification to send.
        :param notification_to_replace: Notification to replace, if any.
        """
        if not self.notifier:
            raise RuntimeError("Could not get toast notifier")

        if notification_to_replace:
            platform_nid = cast(str, notification_to_replace.identifier)
        else:
            platform_nid = str(uuid.uuid4())

        # Create notification XML.
        toast_xml = Element("toast", {"launch": "default"})
        visual_xml = SubElement(toast_xml, "visual")
        actions_xml = SubElement(toast_xml, "actions")

        if notification.thread:
            SubElement(
                toast_xml,
                "header",
                {"id": notification.thread, "title": notification.thread},
            )

        binding = SubElement(visual_xml, "binding", {"template": "ToastGeneric"})

        title_xml = SubElement(binding, "text")
        title_xml.text = notification.title

        message_xml = SubElement(binding, "text")
        message_xml.text = notification.message

        if notification.icon:
            SubElement(
                binding,
                "image",
                {"placement": "appLogoOverride", "src": notification.icon},
            )

        if notification.attachment:
            SubElement(
                binding, "image", {"placement": "hero", "src": notification.attachment}
            )

        if notification.reply_field:
            SubElement(actions_xml, "input", {"id": "textBox", "type": "text"})
            reply_button_xml = SubElement(
                actions_xml,
                "action",
                {
                    "content": notification.reply_field.button_title,
                    "activationType": "background",
                    "arguments": "action=reply&amp",
                },
            )

            # If there are no other buttons, show the
            # reply buttons next to the text field.
            if not notification.buttons:
                reply_button_xml.set("hint-inputId", "textBox")

        for n, button in enumerate(notification.buttons):
            SubElement(
                actions_xml,
                "action",
                {
                    "content": button.title,
                    "activationType": "background",
                    "arguments": str(n),
                },
            )

        if notification.sound:
            SubElement(
                toast_xml, "audio", {"src": "ms-winsoundevent:Notification.Default"}
            )
        else:
            SubElement(toast_xml, "audio", {"silent": "true"})

        xml_document = dom.XmlDocument()
        xml_document.load_xml(tostring(toast_xml, encoding="unicode"))

        native = ToastNotification(xml_document)
        native.tag = platform_nid
        native.priority = self._to_native_urgency[notification.urgency]

        if notification.thread:
            native.group = notification.thread
        else:
            native.group = "default"

        def on_activated(sender, boxed_activated_args) -> None:  # type:ignore
            activated_args = ToastActivatedEventArgs._from(boxed_activated_args)
            action_id = cast(str, activated_args.arguments)

            if action_id == "default":
                if notification.on_clicked:
                    notification.on_clicked()

            elif action_id == "action=reply&amp":
                if notification.reply_field and notification.reply_field.on_replied:
                    boxed_text = activated_args.user_input["textBox"]
                    text = unbox_winrt(boxed_text)
                    notification.reply_field.on_replied(text)

            elif action_id.isnumeric():
                action_number = int(action_id)
                callback = notification.buttons[action_number].on_pressed
                if callback:
                    callback()

        def on_dismissed(sender, dismissed_args) -> None:  # type:ignore
            self._clear_notification_from_cache(notification)

            if dismissed_args.reason == ToastDismissalReason.USER_CANCELED:
                if notification.on_dismissed:
                    notification.on_dismissed()

        def on_failed(sender, failed_args) -> None:  # type:ignore
            logger.warning(
                f"Notification failed (error code {failed_args.error_code.value})"
            )

        native.add_activated(on_activated)
        native.add_dismissed(on_dismissed)
        native.add_failed(on_failed)

        self.notifier.show(native)

        return platform_nid

    async def _clear(self, notification: Notification) -> None:
        """c
        Asynchronously removes a notification from the notification center.
        """
        group = notification.thread or "default"
        self.manager.history.remove(notification.identifier, group, self.app_id)

    async def _clear_all(self) -> None:
        """
        Asynchronously clears all notifications from notification center.
        """
        self.manager.history.clear(self.app_id)


def unbox_winrt(boxed_value: _winrt.Object) -> Any:
    """
    Unbox winrt object. See https://github.com/pywinrt/pywinrt/issues/8.
    """
    if boxed_value is None:
        return boxed_value

    value = IPropertyValue._from(boxed_value)

    if value.type is PropertyType.EMPTY:
        return None
    elif value.type is PropertyType.UINT8:
        return value.get_uint8()
    elif value.type is PropertyType.INT16:
        return value.get_int16()
    elif value.type is PropertyType.UINT16:
        return value.get_uint16()
    elif value.type is PropertyType.STRING:
        return value.get_string()
    else:
        raise NotImplementedError(f"Unboxing {value.type} is not yet supported")



# -*- coding: utf-8 -*-
"""
This module handles desktop notifications and supports multiple backends, depending on
the platform.
"""

from __future__ import annotations

# system imports
import platform
from threading import RLock
import logging
import asyncio
from pathlib import Path
from typing import (
    Type,
    Callable,
    Coroutine,
    List,
    Any,
    TypeVar,
    Sequence,
)

# external imports
from packaging.version import Version

# local imports
from .base import (
    Urgency,
    Button,
    ReplyField,
    Notification,
    DesktopNotifierBase,
    PYTHON_ICON_PATH,
)

__all__ = [
    "Notification",
    "Button",
    "ReplyField",
    "Urgency",
    "DesktopNotifier",
]

logger = logging.getLogger(__name__)

T = TypeVar("T")


default_event_loop_policy = asyncio.DefaultEventLoopPolicy()


def get_implementation() -> Type[DesktopNotifierBase]:
    """
    Return the backend class depending on the platform and version.

    :returns: A desktop notification backend suitable for the current platform.
    :raises RuntimeError: when passing ``macos_legacy = True`` on macOS 12.0 and later.
    """
    if platform.system() == "Darwin":
        from .macos_support import is_bundle, is_signed_bundle, macos_version

        has_unusernotificationcenter = macos_version >= Version("10.14")
        has_nsusernotificationcenter = macos_version < Version("12.0")
        is_signed = is_signed_bundle()

        if has_unusernotificationcenter and is_signed:
            # Use modern UNUserNotificationCenter.
            from .macos import CocoaNotificationCenter

            return CocoaNotificationCenter

        elif has_nsusernotificationcenter and is_bundle():
            if has_unusernotificationcenter and not is_signed:
                logger.warning(
                    "Running outside of a signed Framework or bundle: "
                    "falling back to NSUserNotificationCenter"
                )
            else:
                logger.warning(
                    "Running on macOS 10.13 or earlier: "
                    "falling back to NSUserNotificationCenter"
                )

            # Use deprecated NSUserNotificationCenter.
            from .macos_legacy import CocoaNotificationCenterLegacy

            return CocoaNotificationCenterLegacy

        else:
            # Use dummy backend.
            logger.warning(
                "Notification Center can only be used "
                "from a signed Framework or app bundle"
            )

            from .dummy import DummyNotificationCenter

            return DummyNotificationCenter

    elif platform.system() == "Linux":
        from .dbus import DBusDesktopNotifier

        return DBusDesktopNotifier

    elif platform.system() == "Windows" and Version(platform.version()) >= Version(
        "10.0.10240"
    ):
        from .winrt import WinRTDesktopNotifier

        return WinRTDesktopNotifier

    else:
        from .dummy import DummyNotificationCenter

        return DummyNotificationCenter


class DesktopNotifier:
    """Cross-platform desktop notification emitter

    Uses different backends depending on the platform version and available services.
    All implementations will dispatch notifications without an event loop but will
    require a running event loop to execute callbacks when the end user interacts with a
    notification. On Linux, a asyncio event loop is required. On macOS, a CFRunLoop *in
    the main thread* is required. Packages such as :mod:`rubicon.objc` can be used to
    integrate asyncio with a CFRunLoop.

    :param app_name: Name to identify the application in the notification center. On
        Linux, this should correspond to the application name in a desktop entry. On
        macOS, this argument is ignored and the app is identified by the bundle ID of
        the sending program (e.g., Python).
    :param app_icon: Default icon to use for notifications. This should be either a URI
        string, a :class:`pathlib.Path` path, or a name in a freedesktop.org-compliant
        icon theme. If None, the icon of the calling application will be used if it
        can be determined. On macOS, this argument is ignored and the app icon is
        identified by the bundle ID of the sending program (e.g., Python).
    :param notification_limit: Maximum number of notifications to keep in the system's
        notification center. This may be ignored by some implementations.
    """

    app_icon: str | None

    def __init__(
        self,
        app_name: str = "Python",
        app_icon: Path | str | None = PYTHON_ICON_PATH,
        notification_limit: int | None = None,
    ) -> None:
        impl_cls = get_implementation()

        if isinstance(app_icon, Path):
            self.app_icon = app_icon.as_uri()
        else:
            self.app_icon = app_icon

        self._lock = RLock()
        self._impl = impl_cls(app_name, notification_limit)
        self._did_request_authorisation = False

        # Use our own event loop for the sync API so that we don't interfere with any
        # other ansycio event loops / threads, etc.
        self._loop = default_event_loop_policy.new_event_loop()

    def _run_coro_sync(self, coro: Coroutine[None, None, T]) -> T:
        """
        Runs the given coroutine and returns the result synchronously. This is used as a
        wrapper to conveniently convert the async API calls to synchronous ones.
        """
        if self._loop.is_running():
            future = asyncio.run_coroutine_threadsafe(coro, self._loop)
            res = future.result()
        else:
            res = self._loop.run_until_complete(coro)

        return res

    @property
    def app_name(self) -> str:
        """The application name"""
        return self._impl.app_name

    @app_name.setter
    def app_name(self, value: str) -> None:
        """Setter: app_name"""
        self._impl.app_name = value

    async def request_authorisation(self) -> bool:
        """
        Requests authorisation to send user notifications. This will be automatically
        called for you when sending a notification for the first time. It also can be
        called manually to request authorisation in advance.

        On some platforms such as macOS and iOS, a prompt will be shown to the user
        when this method is called for the first time. This method does nothing on
        platforms where user authorisation is not required.

        :returns: Whether authorisation has been granted.
        """
        with self._lock:
            self._did_request_authorisation = True
            return await self._impl.request_authorisation()

    async def has_authorisation(self) -> bool:
        """Returns whether we have authorisation to send notifications."""
        return await self._impl.has_authorisation()

    async def send(
        self,
        title: str,
        message: str,
        urgency: Urgency = Urgency.Normal,
        icon: Path | str | None = None,
        buttons: Sequence[Button] = (),
        reply_field: ReplyField | None = None,
        on_clicked: Callable[[], Any] | None = None,
        on_dismissed: Callable[[], Any] | None = None,
        attachment: Path | str | None = None,
        sound: bool = False,
        thread: str | None = None,
        timeout: int = -1,
    ) -> Notification:
        """
        Sends a desktop notification.

        Some arguments may be ignored, depending on the backend.

        This method will always return a :class:`base.Notification` instance and will
        not raise an exception when scheduling the notification fails. If the
        notification was scheduled successfully, its ``identifier`` will be set to the
        platform's native notification identifier. Otherwise, the ``identifier`` will be
        ``None``.

        Note that even a successfully scheduled notification may not be displayed to the
        user, depending on their notification center settings (for instance if "do not
        disturb" is enabled on macOS).

        :param title: Notification title.
        :param message: Notification message.
        :param urgency: Notification level: low, normal or critical. This may be
            interpreted differently by some implementations, for instance causing the
            notification to remain visible for longer, or may be ignored.
        :param icon: Optional URI string, :class:`pathlib.Path` or icon name to use. If
            given, this will replace the icon specified by :attr:`app_icon`. Will be
            ignored on macOS.
        :param buttons: A list of buttons with callbacks for the notification.
        :param reply_field: Optional reply field to show with the notification. Can be
            used for instance in chat apps.
        :param on_clicked: Optional callback to call when the notification is clicked.
            The callback will be called without any arguments. This is ignored by some
            implementations.
        :param on_dismissed: Optional callback to call when the notification is
            dismissed. The callback will be called without any arguments. This is
            ignored by some implementations.
        :param attachment: Optional URI string or :class:`pathlib.Path` for an
            attachment to the notification such as an image, movie, or audio file. A
            preview of this may be displayed together with the notification. Different
            platforms and Linux notification servers support different types of
            attachments. Please consult the platform support section of the
            documentation.
        :param sound: Whether to play a sound when the notification is shown. The
            platform's default sound will be used, where available.
        :param thread: An identifier to group related notifications together. This is
            ignored on Linux.
        :param timeout: The duration (in seconds) for which the notification is shown
            unless dismissed. Only supported on Linux. Default is ``-1`` which implies
            OS-specified.

        :returns: The scheduled notification instance.
        """
        if icon is None:
            icon = self.app_icon
        elif isinstance(icon, Path):
            icon = icon.as_uri()

        if isinstance(attachment, Path):
            attachment = attachment.as_uri()

        notification = Notification(
            title,
            message,
            urgency,
            icon,
            buttons,
            reply_field,
            on_clicked,
            on_dismissed,
            attachment,
            sound,
            thread,
            timeout,
        )

        with self._lock:
            # Ask for authorisation if not already done. On some platforms, this will
            # trigger a system dialog to ask the user for permission.
            if not self._did_request_authorisation:
                await self.request_authorisation()

            # We attempt to send the notification regardless of authorization.
            # The user may have changed settings in the meantime.
            await self._impl.send(notification)

            return notification

    def send_sync(
        self,
        title: str,
        message: str,
        urgency: Urgency = Urgency.Normal,
        icon: Path | str | None = None,
        buttons: Sequence[Button] = (),
        reply_field: ReplyField | None = None,
        on_clicked: Callable[[], Any] | None = None,
        on_dismissed: Callable[[], Any] | None = None,
        attachment: Path | str | None = None,
        sound: bool = False,
        thread: str | None = None,
        timeout: int = -1,
    ) -> Notification:
        """
        Synchronous call of :meth:`send`, for use without an asyncio event loop.

        :returns: The scheduled notification instance.
        """
        coro = self.send(
            title,
            message,
            urgency,
            icon,
            buttons,
            reply_field,
            on_clicked,
            on_dismissed,
            attachment,
            sound,
            thread,
            timeout,
        )
        return self._run_coro_sync(coro)

    @property
    def current_notifications(self) -> List[Notification]:
        """A list of all currently displayed notifications for this app"""
        return self._impl.current_notifications

    async def clear(self, notification: Notification) -> None:
        """
        Removes the given notification from the notification center.

        :param notification: Notification to clear.
        """
        with self._lock:
            await self._impl.clear(notification)

    async def clear_all(self) -> None:
        """
        Removes all currently displayed notifications for this app from the notification
        center.
        """
        with self._lock:
            await self._impl.clear_all()
